from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from qtpy import QtCore, QtGui, QtWidgets

from . import utils
from .box import CollapsibleBox
from .rediotab import RadioTabWidget
from .scrollarea import VerticalScrollArea
from .widgets import ParameterWidget, BoolParameter


class Separator(QtWidgets.QWidget):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(QtCore.QMargins(0, 9, 0, 9))
        self.setLayout(layout)

        frame = QtWidgets.QFrame(self)
        frame.setFixedHeight(1)
        frame.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        frame.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        layout.addWidget(frame)


class ParameterToolTip(QtWidgets.QFrame):
    def __init__(
        self, widget: ParameterWidget, parent: QtWidgets.QWidget | None = None
    ) -> None:
        super().__init__(parent)

        self.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        self.setAutoFillBackground(True)

        palette = self.palette()
        palette.setColor(
            QtGui.QPalette.ColorRole.Window,
            palette.color(QtGui.QPalette.ColorRole.Base),
        )
        self.setPalette(palette)

        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setSizeConstraint(QtWidgets.QLayout.SizeConstraint.SetFixedSize)

        title = QtWidgets.QLabel(widget.label(), self)
        font = title.font()
        font.setBold(True)
        title.setFont(font)
        self.layout().addWidget(title)

        separator = QtWidgets.QFrame(self)
        separator.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.layout().addWidget(separator)

        typ = type(widget).__name__.replace('Parameter', '')
        detail = QtWidgets.QLabel(f'Parameter: {widget.name()} ({typ})', self)
        self.layout().addWidget(detail)

        tooltip = QtWidgets.QLabel(widget.tooltip(), self)
        # tooltip.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        tooltip.setWordWrap(True)
        tooltip.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignTop | QtCore.Qt.AlignmentFlag.AlignLeft
        )
        self.layout().addWidget(tooltip)

    def focusOutEvent(self, event: QtCore.QEvent) -> None:
        self.hide()

    def leaveEvent(self, event: QtCore.QEvent) -> None:
        self.hide()


class ParameterLabel(QtWidgets.QLabel):
    def __init__(
        self, widget: ParameterWidget, parent: QtWidgets.QWidget | None = None
    ) -> None:
        super().__init__(widget.label(), parent)

        self._tooltip: ParameterToolTip | None = None
        self._widget = widget

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.text())})'

    def enterEvent(self, event: QtGui.QEnterEvent) -> None:
        if self._widget.tooltip():
            QtCore.QTimer.singleShot(600, self.show_tooltip)
        super().enterEvent(event)

    def show_tooltip(self) -> None:
        global_position = QtGui.QCursor.pos()
        if self.geometry().contains(self.parent().mapFromGlobal(global_position)):
            if self._tooltip is None:
                self._tooltip = ParameterToolTip(self._widget)
                self._tooltip.setParent(self.window(), QtCore.Qt.WindowType.ToolTip)
            self._tooltip.move(global_position)
            self._tooltip.show()


class LabelFilter(QtCore.QObject):
    def __init__(
        self, label: QtWidgets.QLabel, parent: QtWidgets.QWidget | None = None
    ) -> None:
        if not parent:
            # Store the LabelFilter on the label, as it should be deleted with the
            # label and not the watched widget.
            parent = label
        super().__init__(parent)
        self.label = label

    def eventFilter(self, watched: QtCore.QObject, event: QtCore.QEvent) -> bool:
        if isinstance(watched, QtWidgets.QWidget):
            if event.type() == QtCore.QEvent.Type.EnabledChange:
                self.label.setEnabled(watched.isEnabled())
            elif event.type() in (QtCore.QEvent.Type.Show, QtCore.QEvent.Type.Hide):
                self.label.setVisible(watched.isVisible())
        return False


class ParameterForm(QtWidgets.QWidget):
    parameter_changed: QtCore.Signal = QtCore.Signal(ParameterWidget)

    def __init__(self, name: str = '', parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._widgets: dict[str, QtWidgets.QWidget] = {}
        self._name = name
        self._root = self
        self._flat = False
        self._unique_names = False

        self._layout = QtWidgets.QGridLayout()
        self.setLayout(self._layout)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._name!r})'

    def name(self) -> str:
        return self._name

    def set_name(self, name: str) -> None:
        self._name = name

    def flat(self) -> bool:
        return self._flat

    def set_flat(self, flat: bool) -> None:
        """
        If True, the form doesn't create a hierarchy when querying parameters or values.
        """
        self._flat = flat

    def unique_names(self) -> bool:
        return self._unique_names

    def set_unique_names(self, unique_names: bool) -> None:
        """
        If True, requires that all parameters have a unique name in the root's
        hierarchy.
        """
        self._unique_names = unique_names

    def root(self) -> ParameterForm:
        return self._root

    def set_root(self, root: ParameterForm) -> None:
        """
        Set the root which is used when checking for unique names in the hierarchy.
        """
        self._root = root

    def state(self) -> dict:
        """Return the state of the form as a dict."""

        state = {}
        if boxes := self._collapsed_boxes():
            state['collapsed_boxes'] = boxes
        for form in self.forms():
            if form_state := form.state():
                state[form.name()] = form_state
        return state

    def set_state(self, state: dict) -> None:
        """Load the state of the form from a dict."""

        boxes = state.get('collapsed_boxes')
        if boxes is not None:
            self._set_collapsed_boxes(boxes)
        for form in self.forms():
            if form_state := state.get(form.name()):
                form.set_state(form_state)

    def values(self) -> dict[str, Any]:
        """Return a nested dict of all values."""

        widgets = self.widgets()
        values = {}
        for name, widget in widgets.items():
            if isinstance(widget, ParameterWidget):
                values[name] = widget.value()
            elif isinstance(widget, ParameterForm):
                values[name] = widget.values()
            elif isinstance(widget, CollapsibleBox):
                values[name] = widget.checked()
            elif isinstance(widget, RadioTabWidget):
                if form := self.form(name.replace('_enabled', '')):
                    values[name] = widget.currentWidget() == form
        return values

    def set_values(self, values: dict) -> None:
        """Set the values of ParameterWidgets in the form."""

        widgets = self.widgets()
        for name, value in values.items():
            widget = widgets.get(name)
            if isinstance(widget, ParameterWidget):
                widget.set_value(value)
            elif isinstance(widget, ParameterForm):
                widget.set_values(value)
            elif isinstance(widget, CollapsibleBox):
                widget.set_checked(value)
            elif isinstance(widget, RadioTabWidget):
                if value:
                    if form := self.form(name.replace('_enabled', '')):
                        widget.setCurrentWidget(form)

    def defaults(self) -> dict[str, Any]:
        """Return a nested dict of all defaults."""

        widgets = self.widgets()
        defaults = {}
        for name, widget in widgets.items():
            if isinstance(widget, ParameterWidget):
                defaults[name] = widget.default()
            elif isinstance(widget, ParameterForm):
                defaults[name] = widget.defaults()
            elif isinstance(widget, RadioTabWidget):
                if form := self.form(name.replace('_enabled', '')):
                    defaults[name] = widget.currentWidget() == form
        return defaults


    def set_defaults(self, values: dict) -> None:
        """Set the default values of ParameterWidgets in the form."""

        widgets = self.widgets()
        for name, value in values.items():
            widget = widgets.get(name)
            if isinstance(widget, ParameterWidget):
                widget.set_default(value)
            elif isinstance(widget, ParameterForm):
                widget.set_defaults(value)

    def reset(self) -> None:
        """Reset all parameters in this form to their default values."""

        widgets = self.widgets()
        for name, widget in widgets.items():
            if isinstance(widget, (ParameterWidget, ParameterForm)):
                widget.reset()
            elif isinstance(widget, CollapsibleBox):
                widget.set_checked(True)

    def add_parameter(
        self,
        widget: ParameterWidget,
        checkable: bool = False,
        alignment: QtCore.Qt.AlignmentFlag | None = None,
    ) -> ParameterWidget:
        """Add a parameter to the ParameterForm's GridLayout."""

        name = widget.name()
        self._validate_name(name)
        self._widgets[name] = widget

        row = self._layout.rowCount() - 1

        # Checkbox
        if checkable:
            checkbox_name = f'{name}_enabled'
            checkbox = BoolParameter(checkbox_name)
            self._layout.addWidget(checkbox, row, 0)
            checkbox.set_value(False)
            widget.blockSignals(True)
            widget.setEnabled(False)
            widget.blockSignals(False)
            checkbox.value_changed.connect(widget.setEnabled)
            checkbox.value_changed.connect(
                lambda: self.parameter_changed.emit(checkbox)
            )

            self._widgets[checkbox_name] = checkbox

        # Label
        if widget.label():
            label = ParameterLabel(widget, self)
            label.setEnabled(widget.isEnabled())
            self._layout.addWidget(label, row, 1)
            if alignment:
                label.setAlignment(alignment)
            label.setVisible(False)
            label_filter = LabelFilter(label)
            widget.installEventFilter(label_filter)

        # Widget
        self._layout.addWidget(widget, row, 2)
        widget.value_changed.connect(lambda: self.parameter_changed.emit(widget))

        self._refresh_stretch()
        return widget

    def add_form(self, form: ParameterForm, checkable: bool = False) -> CollapsibleBox:
        """Add a form and return the CollapsibleBox."""

        # Add ParameterForm
        name = form.name()
        self._validate_name(name)
        self._widgets[name] = form
        form.set_root(self)
        form.parameter_changed.connect(self.parameter_changed)

        # Add CollapsibleBox
        label = utils.title(name)
        box = CollapsibleBox(label)
        box.set_collapsible(True)
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(QtCore.QMargins())
        layout.setSpacing(0)
        layout.addWidget(form)
        box.setLayout(layout)

        if checkable:
            box.set_checkable(checkable)
            self._widgets[f'{name}_enabled'] = box

        self.add_widget(box)

        return box

    def add_forms(
        self, forms: Sequence[ParameterForm], radio: bool = False
    ) -> QtWidgets.QTabWidget:
        """
        Add multiple forms and return the TabWidget. If radio is True, use a
        RadioTabWidget which only allows one form to be active.
        """

        # Validate the names before creating any widgets.
        for form in forms:
            self._validate_name(form.name())

        if radio:
            tab_widget = RadioTabWidget()
        else:
            tab_widget = QtWidgets.QTabWidget()

        for form in forms:
            name = form.name()
            self._widgets[name] = form
            form.set_root(self)
            form.parameter_changed.connect(self.parameter_changed)

            label = utils.title(name)
            tab_widget.addTab(form, label)

            if radio:
                self._widgets[f'{name}_enabled'] = tab_widget

        self.add_widget(tab_widget)

        return tab_widget

    def add_widget(
        self,
        widget: QtWidgets.QWidget,
        column: int = 1,
        column_span: int = -1,
        row: int = -1,
        row_span: int = 1,
    ) -> QtWidgets.QWidget:
        """Add a widget to the ParameterForm's GridLayout."""

        if row < 0:
            row = self._layout.rowCount() + row
        if column_span == -1 and self._layout.columnCount() < 2:
            # In case the layout doesn't have any columns yet, default to 3 columns.
            column_span = 2
        self._layout.addWidget(widget, row, column, row_span, column_span)
        self._refresh_stretch()
        return widget

    def add_separator(
        self,
        name: str = '',
        column: int = 2,
        column_span: int = -1,
        row: int = -1,
        row_span: int = 1,
    ) -> Separator:
        """Add a horizontal separator to the ParameterForm's GridLayout."""

        if row < 0:
            row = self._layout.rowCount() + row
        if name:
            label = QtWidgets.QLabel(name)
            self._layout.addWidget(label, row, 1)
            column = min(2, column)
        separator = Separator()
        self.add_widget(separator, column, column_span, row, row_span)
        return separator

    def add_layout(
        self,
        layout: QtWidgets.QLayout,
        column: int = 1,
        column_span: int = -1,
        row: int = -1,
        row_span: int = 1,
    ) -> QtWidgets.QLayout:
        """Add a layout to the ParameterForm's GridLayout."""

        if row < 0:
            row = self._layout.rowCount() + row
        if column_span == -1 and self._layout.columnCount() < 2:
            # In case the layout doesn't have any columns yet, default to 3 columns.
            column_span = 2
        self._layout.addLayout(layout, row, column, row_span, column_span)
        self._refresh_stretch()
        return layout

    def clear(self) -> None:
        """Clear the form from all parameters, widgets and forms."""

        for i in reversed(range(self._layout.count())):
            if item := self._layout.itemAt(i):
                if widget := item.widget():
                    self._layout.removeWidget(widget)
                    widget.deleteLater()
                else:
                    self._layout.removeItem(item)
        self._widgets = {}

    def remove_parameter(self, parameter: ParameterWidget) -> None:
        """Remove and delete a parameter from the form."""

        index = self._layout.indexOf(parameter)
        if index < 0:
            return
        row, column, row_span, col_span = self._layout.getItemPosition(index)
        for i in range(3):
            if item := self._layout.itemAtPosition(row, i):
                if widget := item.widget():
                    self._layout.removeWidget(widget)
                    widget.deleteLater()
        self._refresh_stretch()

    def remove_form(self, form: ParameterForm) -> None:
        """Remove a form from the form. This does not remove box or tab widgets."""
        name = form.name()
        if name in self._widgets:
            del self._widgets[name]
        box_name = f'{name}_enabled'
        if box_name in self._widgets:
            del self._widgets[box_name]

    def remove_widget(self, widget: QtWidgets.QWidget) -> None:
        """Remove and delete a widget from the form."""

        self._layout.removeWidget(widget)
        widget.deleteLater()
        self._refresh_stretch()

    def parameter(self, name: str) -> ParameterWidget | None:
        """
        Return the first ParameterWidget with name `name`.

        The `name` follows the attribute naming scheme, meaning nested parameters can
        be accessed with: name = 'parent.child'.
        """

        names = name.split('.')

        widgets = self.widgets()
        for n in names:
            widget = widgets.get(n)
            if isinstance(widget, ParameterForm):
                widgets = widget.widgets()
            elif isinstance(widget, ParameterWidget):
                return widget
            else:
                break
        return None

    def parameters(self) -> tuple[ParameterWidget, ...]:
        """Return all ParameterWidgets of this form."""

        widgets = self.widgets().values()
        parameters = tuple(w for w in widgets if isinstance(w, ParameterWidget))
        return parameters

    def form(self, name: str) -> ParameterForm | None:
        """
        Return the first ParameterForm with name `name`.

        The `name` follows the attribute naming scheme, meaning nested forms can
        be accessed with: name = 'parent.child'.
        """

        names = name.split('.')

        widgets = self.widgets()
        for n in names:
            widget = widgets.get(n)
            if isinstance(widget, ParameterForm):
                if n == names[-1]:
                    return widget
                else:
                    widgets = widget.widgets()
            else:
                break
        return None

    def forms(self) -> tuple[ParameterForm, ...]:
        """Return all ParameterForms of this form."""

        widgets = self.widgets().values()
        forms = tuple(w for w in widgets if isinstance(w, ParameterForm))
        return forms

    def boxes(self) -> tuple[CollapsibleBox, ...]:
        """Return all CollapsibleBoxes of this form."""

        boxes = []
        for i in reversed(range(self._layout.count())):
            if item := self._layout.itemAt(i):
                if widget := item.widget():
                    if isinstance(widget, CollapsibleBox):
                        boxes.append(widget)
        return tuple(boxes)

    def widgets(self) -> dict:
        """
        Return a dict with all parameter widgets of the form and its children.
        The dict will be flattened if the child forms are flat.
        """

        widgets = {}
        for name, widget in self._widgets.items():
            if isinstance(widget, ParameterForm):
                if widget.flat():
                    widgets.update(widget.widgets())
                else:
                    widgets[name] = widget
            elif isinstance(widget, ParameterWidget):
                widgets[name] = widget
            elif isinstance(widget, CollapsibleBox):
                widgets[name] = widget
            elif isinstance(widget, RadioTabWidget):
                widgets[name] = widget
        return widgets

    @staticmethod
    def checkbox(parameter: ParameterWidget) -> BoolParameter | None:
        """Return the BoolParameter for a ParameterWidget `parameter` if it exists."""

        if parent := parameter.parentWidget():
            layout = parent.layout()
            if layout and isinstance(layout, QtWidgets.QGridLayout):
                index = layout.indexOf(parameter)
                if index >= 0:
                    row, column, row_span, col_span = layout.getItemPosition(index)
                    if item := layout.itemAtPosition(row, 0):
                        if widget := item.widget():
                            if isinstance(widget, BoolParameter):
                                return widget
        return None

    @staticmethod
    def label(parameter: ParameterWidget) -> ParameterLabel | None:
        """Return the ParameterLabel for a ParameterWidget `parameter` if it exists."""

        if parent := parameter.parentWidget():
            layout = parent.layout()
            if layout and isinstance(layout, QtWidgets.QGridLayout):
                index = layout.indexOf(parameter)
                if index >= 0:
                    row, column, row_span, col_span = layout.getItemPosition(index)
                    if item := layout.itemAtPosition(row, 1):
                        if widget := item.widget():
                            if isinstance(widget, ParameterLabel):
                                return widget
        return None

    def _collapsed_boxes(self) -> tuple[str, ...]:
        """Return a tuple of all collapsed boxes."""

        boxes = []
        for box in self.boxes():
            if box.collapsed():
                boxes.append(box.title())
        return tuple(boxes)

    def _set_collapsed_boxes(self, boxes: Sequence[str]) -> None:
        """Collapse the boxes with title in `boxes`."""

        for box in self.boxes():
            if box.collapsible():
                box.set_collapsed(box.title() in boxes)

    def _refresh_stretch(self) -> None:
        """Ensure the last column and row is expanding."""

        self._layout.setColumnStretch(self._layout.columnCount() - 1, 1)
        self._layout.setRowStretch(self._layout.rowCount() - 1, 0)
        self._layout.setRowStretch(self._layout.rowCount(), 1)

    def _names(self) -> tuple[str, ...]:
        """Return a flat tuple of all child parameter's names."""

        names = []
        for name, widget in self._widgets.items():
            if isinstance(widget, ParameterForm):
                names.extend(widget._names())
            else:
                names.append(name)
        return tuple(names)

    def _validate_name(self, name: str) -> None:
        """
        Validates a parameter name, ensuring it does not already exist.

        :raises ValueError: If the name is not valid.
        """
        if not name:
            raise ValueError(f'name cannot be empty')

        if self._root._unique_names:
            names = self._root._names()
        else:
            names = self._widgets.keys()

        if name in names:
            raise ValueError(f'name {name!r} is not unique')


class ParameterEditor(ParameterForm):
    """ParameterEditor is a ParameterForm in a VerticalScrollArea."""

    def __init__(self, name: str = '', parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(name=name, parent=parent)

        # Move the GridLayout to another widget first
        widget = QtWidgets.QWidget()
        widget.setLayout(self._layout)

        # Create the ScrollArea and add the widget
        self.scroll_area = VerticalScrollArea()
        self.scroll_area.setWidget(widget)
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(QtCore.QMargins())
        layout.addWidget(self.scroll_area)
        self.setLayout(layout)

    def sizeHint(self) -> QtCore.QSize:
        return self.scroll_area.sizeHint()



from __future__ import annotations

from qt_material_icons import MaterialIcon
from qtpy import QtWidgets


class RadioTabWidget(QtWidgets.QTabWidget):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._unchecked_icon = MaterialIcon('radio_button_unchecked')
        self._checked_icon = MaterialIcon('radio_button_checked')

        self.currentChanged.connect(self._current_changed)

    def tabInserted(self, index: int) -> None:
        if not self.tabIcon(index):
            self.setTabIcon(index, self._unchecked_icon)

    def _current_changed(self, index: int) -> None:
        for i in range(self.count()):
            if i == index:
                self.setTabIcon(i, self._checked_icon)
            else:
                self.setTabIcon(i, self._unchecked_icon)



from __future__ import annotations

import numbers
import typing
from collections.abc import Sequence

from qt_material_icons import MaterialIcon
from qtpy import QtCore, QtGui, QtWidgets

from .resizegrip import ResizeGrip
from .widgets import FloatParameter, IntParameter, ParameterWidget


class StyledItemDelegate(QtWidgets.QStyledItemDelegate):
    def setModelData(
        self,
        editor: QtWidgets.QWidget,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ) -> None:
        indexes = self.selected_indexes(index)
        model.blockSignals(True)
        for index in indexes:
            if index == indexes[-1]:
                model.blockSignals(False)
            super().setModelData(editor, model, index)

    def selected_indexes(self, current_index: QtCore.QModelIndex | None):
        indexes = []
        if (parent := self.parent()) and isinstance(parent, QtWidgets.QTreeView):
            indexes = parent.selectedIndexes()
        if current_index is not None and current_index not in indexes:
            indexes.append(current_index)
        return indexes

    def set_edit_data(
        self,
        value: typing.Any,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ) -> None:
        indexes = self.selected_indexes(index)
        model.blockSignals(True)
        for index in indexes:
            if index == indexes[-1]:
                model.blockSignals(False)
            model.setData(index, value, QtCore.Qt.ItemDataRole.EditRole)


class IntegerDelegate(StyledItemDelegate):
    def displayText(self, value: typing.Any, locale: QtCore.QLocale) -> str:
        return str(value)

    def createEditor(
        self,
        parent: QtWidgets.QWidget,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ) -> QtWidgets.QWidget:
        editor = IntParameter(parent=parent)
        editor.set_slider_visible(False)
        editor.set_commit_on_edit(True)
        editor.line.setFrame(False)
        return editor

    def setEditorData(
        self, editor: QtWidgets.QWidget, index: QtCore.QModelIndex
    ) -> None:
        value = index.model().data(index, QtCore.Qt.ItemDataRole.EditRole)
        if value and isinstance(editor, IntParameter):
            editor.set_value(value)

    def setModelData(
        self,
        editor: QtWidgets.QWidget,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ) -> None:
        if isinstance(editor, IntParameter):
            value = editor.value()
            self.set_edit_data(value, model, index)

    def updateEditorGeometry(
        self,
        editor: QtWidgets.QWidget,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ) -> None:
        editor.setGeometry(option.rect)


class FloatDelegate(StyledItemDelegate):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)
        self.decimals = None

    def displayText(self, value: typing.Any, locale: QtCore.QLocale) -> str:
        if self.decimals is not None:
            return f'{value:.{self.decimals}f}'.rstrip('0').rstrip('.')
        else:
            return f'{value}'.rstrip('0').rstrip('.')

    def createEditor(
        self,
        parent: QtWidgets.QWidget,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ) -> QtWidgets.QWidget:
        editor = FloatParameter(parent=parent)
        editor.set_slider_visible(False)
        editor.set_decimals(6)
        editor.set_commit_on_edit(True)
        editor.line.setFrame(False)
        return editor

    def setEditorData(
        self, editor: QtWidgets.QWidget, index: QtCore.QModelIndex
    ) -> None:
        value = index.model().data(index, QtCore.Qt.ItemDataRole.EditRole)
        if value and isinstance(editor, FloatParameter):
            editor.set_value(value)

    def setModelData(
        self,
        editor: QtWidgets.QWidget,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ) -> None:
        if isinstance(editor, FloatParameter):
            value = editor.value()
            self.set_edit_data(value, model, index)

    def updateEditorGeometry(
        self,
        editor: QtWidgets.QWidget,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ) -> None:
        editor.setGeometry(option.rect)


class DataTableModel(QtGui.QStandardItemModel):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self.types = []

    def setData(
        self,
        index: QtCore.QModelIndex,
        value: typing.Any,
        role: int = QtCore.Qt.ItemDataRole.EditRole,
    ) -> bool:
        if role == QtCore.Qt.ItemDataRole.EditRole:
            column = index.column()
            typ = self.types[column] if column < len(self.types) else None
            if typ is not None:
                try:
                    value = typ(value)
                except (ValueError, TypeError):
                    value = None
        return super().setData(index, value, role)


class DataTableView(QtWidgets.QTableView):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._init_ui()

    def _init_ui(self) -> None:
        mode = QtWidgets.QAbstractItemView.SelectionMode.ContiguousSelection
        self.setSelectionMode(mode)
        self.setAlternatingRowColors(True)
        self.setShowGrid(False)
        self.setSortingEnabled(True)

        self.horizontalHeader().setSortIndicatorShown(True)
        self.horizontalHeader().setSectionsMovable(True)
        self.horizontalHeader().setStretchLastSection(True)

        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum
        )

        # Context menu
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_menu)
        self.context_menu = QtWidgets.QMenu(self)

        action = QtGui.QAction("Edit", self)
        action.triggered.connect(self.edit_selected)
        self.addAction(action)
        self.context_menu.addAction(action)

        action = QtGui.QAction("Copy", self)
        action.setShortcut(QtGui.QKeySequence.StandardKey.Copy)
        action.setShortcutContext(QtCore.Qt.ShortcutContext.WidgetShortcut)
        action.triggered.connect(self.copy_selected)
        self.addAction(action)
        self.context_menu.addAction(action)

        action = QtGui.QAction("Paste", self)
        action.setShortcut(QtGui.QKeySequence.StandardKey.Paste)
        action.setShortcutContext(QtCore.Qt.ShortcutContext.WidgetShortcut)
        action.triggered.connect(self.paste_selected)
        self.addAction(action)
        self.context_menu.addAction(action)

    def copy_selected(self) -> None:
        selected_indexes = self.selectedIndexes()
        if not self.model():
            return

        # Create a nested list from the data
        data = []
        row_data = []
        row = 0
        for index in selected_indexes:
            if row != index.row() and row_data:
                data.append(row_data)
                row_data = []
            row = index.row()
            row_data.append(self.model().data(index))
        if row_data:
            data.append(row_data)

        # Copy to clipboard
        text = '\n'.join('\t'.join(str(d) for d in row_data) for row_data in data)
        clipboard = QtGui.QClipboard()
        clipboard.setText(text)

    def edit_selected(self) -> None:
        index = self.currentIndex()
        if index.isValid():
            self.edit(index)

    def paste_selected(self) -> None:
        selected_indexes = self.selectedIndexes()
        current_index = self.currentIndex()
        if not current_index and not selected_indexes or not self.model():
            return

        # Get top left index
        for index in selected_indexes:
            if (
                index.row() <= current_index.row()
                and index.column() <= current_index.column()
            ):
                current_index = index

        # Get data
        text = QtGui.QClipboard().text()
        data = (row_text.split('\t') for row_text in text.split('\n'))

        # Set data
        for row, row_data in enumerate(data):
            row_index = current_index.siblingAtRow(current_index.row() + row)
            if not row_index.isValid():
                continue
            for column, d in enumerate(row_data):
                index = row_index.siblingAtColumn(current_index.column() + column)
                if not index.isValid():
                    continue
                self.model().setData(index, d)

    def show_menu(self, position: QtCore.QPoint) -> None:
        self.context_menu.exec_(self.viewport().mapToGlobal(position))


class TabDataParameter(ParameterWidget):
    """Parameter to display tabular data in a QTreeWidget."""

    value_changed: QtCore.Signal = QtCore.Signal(tuple)

    _value: tuple = ()
    _default: tuple = ()
    _headers: tuple[str, ...] = ()
    _types: tuple[type, ...] = ()
    _start_index: int = 0
    _decimals: int = 3

    def __init__(
        self, name: str | None = None, parent: QtWidgets.QWidget | None = None
    ) -> None:
        self._delegates = []
        super().__init__(name, parent)

    def _init_ui(self) -> None:
        QtWidgets.QWidget().setLayout(self.layout())
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        # TableView
        self.model = DataTableModel(parent=self)
        self.model.itemChanged.connect(self._item_change)
        self.view = DataTableView(parent=self)
        self.view.setModel(self.model)
        self.view.setSizeAdjustPolicy(
            QtWidgets.QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents
        )
        self.layout().addWidget(self.view)

        # Toolbar
        self.toolbar = QtWidgets.QToolBar()
        size = self.toolbar.style().pixelMetric(
            QtWidgets.QStyle.PixelMetric.PM_SmallIconSize
        )
        self.toolbar.setIconSize(QtCore.QSize(size, size))
        self.layout().addWidget(self.toolbar)

        icon = MaterialIcon('add')
        action = QtGui.QAction(icon, 'Add Row', self)
        action.triggered.connect(self.add_row)
        self.toolbar.addAction(action)

        icon = MaterialIcon('remove')
        action = QtGui.QAction(icon, 'Remove Row', self)
        action.triggered.connect(self.remove_row)
        self.toolbar.addAction(action)

        icon = MaterialIcon('delete')
        action = QtGui.QAction(icon, 'Clear', self)
        action.triggered.connect(self.clear)
        self.toolbar.addAction(action)

        self.resize_grip = ResizeGrip(self.view)
        self.resize_grip.can_resize_horizontal = True

    def clear(self) -> None:
        self.model.clear()
        super().set_value(())
        self._refresh_horizontal_headers()

    def decimals(self) -> int:
        return self._decimals

    def set_decimals(self, decimals: int) -> None:
        self._decimals = decimals
        for delegate in self._delegates:
            if isinstance(delegate, FloatDelegate):
                delegate.decimals = decimals

    def headers(self) -> tuple[str, ...]:
        return self._headers

    def set_headers(self, headers: Sequence[str]) -> None:
        if not isinstance(headers, tuple):
            headers = tuple(headers)
        self._headers = headers
        self._refresh_horizontal_headers()

    def types(self) -> tuple[type, ...]:
        return self._types

    def set_types(self, types: Sequence[type]) -> None:
        if not isinstance(types, tuple):
            types = tuple(types)
        self._types = types

        self._delegates = []
        # Fill up types to column count
        if not types:
            types = [None] * (self.model.columnCount() - len(types))
        self.model.types = types

        for i, type_ in enumerate(types):
            if issubclass(type_, float):
                delegate = FloatDelegate(self.view)
                delegate.decimals = self._decimals
            elif issubclass(type_, int):
                delegate = IntegerDelegate(self.view)
            else:
                delegate = StyledItemDelegate(self.view)
            self.view.setItemDelegateForColumn(i, delegate)
            self._delegates.append(delegate)

    def start_index(self) -> int:
        return self._start_index

    def set_start_index(self, start_index: int) -> None:
        self._start_index = start_index
        self._refresh_vertical_headers()

    def value(self) -> tuple:
        return super().value()

    def set_value(self, value: Sequence) -> None:
        self.model.clear()
        if not value:
            return

        for row, row_data in enumerate(value):
            items = []
            if isinstance(row_data, dict):
                row_data = row_data.values()
            for column, cell_data in enumerate(row_data):
                item = QtGui.QStandardItem()
                item.setData(cell_data, QtCore.Qt.ItemDataRole.EditRole)
                items.append(item)
            if items:
                self.model.appendRow(items)

        super().set_value(self._tab_data_value())
        self._refresh_horizontal_headers()
        self._refresh_vertical_headers()

    def add_row(self) -> None:
        """Add a row."""

        items = []
        for i in range(self.model.columnCount()):
            item = QtGui.QStandardItem()
            if i < len(self._types):
                type_ = self._types[i]
                if issubclass(type_, numbers.Number):
                    item.setData(0, QtCore.Qt.ItemDataRole.EditRole)
                elif issubclass(type_, str):
                    item.setData('', QtCore.Qt.ItemDataRole.EditRole)
            items.append(item)
        self.model.insertRow(self.model.rowCount(), items)
        self._item_change()
        self._refresh_vertical_headers()

    def remove_row(self) -> None:
        """Remove a row."""

        self.model.takeRow(self.model.rowCount() - 1)
        self._item_change()

    def _item_change(self) -> None:
        super().set_value(self._tab_data_value())

    def _tab_data_value(self) -> tuple:
        """Return the data as nested tuples."""

        values = []
        for row in range(self.model.rowCount()):
            row_values = []
            for column in range(self.model.columnCount()):
                index = self.model.index(row, column)
                column_value = self.model.data(index, QtCore.Qt.ItemDataRole.EditRole)
                row_values.append(column_value)
            values.append(tuple(row_values))
        return tuple(values)

    def _refresh_horizontal_headers(self) -> None:
        """Refresh the horizontal header labels."""

        if self._headers:
            labels = self._headers
        else:
            labels = tuple(str(i) for i in range(self.model.columnCount()))
        self.model.setHorizontalHeaderLabels(labels)

        # Resize columns
        header = self.view.horizontalHeader()
        for i in range(header.count()):
            size = max(self.view.sizeHintForColumn(i), header.sectionSizeHint(i))
            self.view.setColumnWidth(i, size)

    def _refresh_vertical_headers(self) -> None:
        """Refresh the vertical header labels."""

        start = self._start_index
        end = self.model.rowCount() + self._start_index
        labels = tuple(str(i) for i in range(start, end))
        self.model.setVerticalHeaderLabels(labels)



from __future__ import annotations

import enum
from collections.abc import Sequence

from qt_material_icons import MaterialIcon
from qtpy import QtCore, QtGui, QtWidgets

PixelMetric = QtWidgets.QStyle.PixelMetric

QStyleOptionTab = QtWidgets.QStyleOptionTab


class CollapsibleBox(QtWidgets.QFrame):
    class Style(enum.IntEnum):
        NONE = 0
        SIMPLE = 1
        BUTTON = 2

    NONE = Style.NONE
    SIMPLE = Style.SIMPLE
    BUTTON = Style.BUTTON

    def __init__(
        self, title: str = '', parent: QtWidgets.QWidget | None = None
    ) -> None:
        super().__init__(parent)

        self._maximum_height = self.maximumHeight()
        self._collapsed = False
        self._checkable = False
        self._collapsible = False
        self._style = CollapsibleBox.Style.SIMPLE

        self.header = None
        self.title_label = None
        self.menu_button = None
        self.frame = None
        self._init_ui()

        if title:
            self.set_title(title)

        self.set_box_style(self._style)

    def _init_ui(self) -> None:
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Maximum
        )

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setSpacing(0)
        super().setLayout(self._layout)

        # Header
        self.header = QtWidgets.QWidget()
        self.header.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)
        self.header.installEventFilter(self)
        self.header.setBackgroundRole(QtGui.QPalette.ColorRole.Button)
        self._layout.addWidget(self.header)

        header_layout = QtWidgets.QHBoxLayout()
        self.header.setLayout(header_layout)
        self._default_margins = header_layout.contentsMargins()

        # Expand
        size = self.style().pixelMetric(PixelMetric.PM_ButtonIconSize)
        self._expand_more_label = QtWidgets.QLabel(self.header)
        self._expand_more_label.setVisible(False)
        icon = MaterialIcon('chevron_right')
        self._expand_more_label.setPixmap(icon.pixmap(size))
        header_layout.addWidget(self._expand_more_label)

        self._expand_less_label = QtWidgets.QLabel(self.header)
        self._expand_less_label.setVisible(False)
        icon = MaterialIcon('expand_more')
        self._expand_less_label.setPixmap(icon.pixmap(size))
        header_layout.addWidget(self._expand_less_label)

        # Checkbox
        self.checkbox = QtWidgets.QCheckBox(self.header)
        self.checkbox.setVisible(False)
        header_layout.addWidget(self.checkbox)

        # Title
        self.title_label = QtWidgets.QLabel(self.header)
        header_layout.addWidget(self.title_label)
        header_layout.addStretch()

        # Menu
        self.menu_button = QtWidgets.QToolButton(self.header)
        self.menu_button.setIcon(MaterialIcon('menu'))
        self.menu_button.setAutoRaise(True)
        self.menu_button.setVisible(False)
        self.menu_button.setPopupMode(
            QtWidgets.QToolButton.ToolButtonPopupMode.InstantPopup
        )
        self.menu_button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self.menu_button.pressed.connect(self._show_menu)
        size = self.header.style().pixelMetric(PixelMetric.PM_SmallIconSize)
        self.menu_button.setMaximumSize(QtCore.QSize(size, size))
        header_layout.addWidget(self.menu_button)

        # Frame
        self.frame = QtWidgets.QFrame(self)
        self._layout.addWidget(self.frame)
        self._layout.setStretch(1, 1)
        self.checkbox.toggled.connect(self.frame.setEnabled)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.title()!r})'

    def actionEvent(self, event: QtGui.QActionEvent) -> None:
        self.menu_button.setVisible(bool(self.actions()))
        super().actionEvent(event)

    def enterEvent(self, event: QtGui.QEnterEvent) -> None:
        self.update()
        super().enterEvent(event)

    def eventFilter(self, watched: QtCore.QObject, event: QtCore.QEvent) -> bool:
        if watched == self.header and self._collapsible:
            if event.type() == QtCore.QEvent.Type.MouseButtonPress:
                event: QtCore.QEvent.Type.MouseButtonPress
                if event.button() == QtCore.Qt.MouseButton.LeftButton:
                    self.header.setAutoFillBackground(True)

            if event.type() == QtCore.QEvent.Type.MouseButtonRelease:
                event: QtCore.QEvent.Type.MouseButtonRelease
                if event.button() == QtCore.Qt.MouseButton.LeftButton:
                    self.set_collapsed(not self._collapsed)
                    self.header.setAutoFillBackground(False)
        return super().eventFilter(watched, event)

    def leaveEvent(self, event: QtCore.QEvent) -> None:
        self.update()
        super().leaveEvent(event)

    def layout(self) -> QtWidgets.QLayout:
        return self.frame.layout()

    def setLayout(self, layout: QtWidgets.QLayout) -> None:
        self.frame.setLayout(layout)

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        if self._style != CollapsibleBox.Style.BUTTON:
            super().paintEvent(event)
            return

        painter = QtGui.QPainter(self)
        style = self.style()

        if self._collapsed:
            option = QtWidgets.QStyleOptionButton()
            option.initFrom(self)
            if self.underMouse():
                option.state |= QtWidgets.QStyle.StateFlag.State_MouseOver

            element = QtWidgets.QStyle.PrimitiveElement.PE_PanelButtonCommand
            style.drawPrimitive(element, option, painter, self)
        else:
            overlap = style.pixelMetric(
                QtWidgets.QStyle.PixelMetric.PM_TabBarBaseOverlap
            )

            # Background Frame
            option = QtWidgets.QStyleOptionTabWidgetFrame()
            option.initFrom(self)
            option.rect.adjust(0, overlap, 0, 0)

            element = QtWidgets.QStyle.PrimitiveElement.PE_FrameTabWidget
            style.drawPrimitive(element, option, painter, self)

            # Header
            option = QtWidgets.QStyleOptionTab()
            option.initFrom(self)

            option.features |= QStyleOptionTab.TabFeature.HasFrame
            option.state |= QtWidgets.QStyle.StateFlag.State_Selected
            option.selectedPosition = QStyleOptionTab.SelectedPosition.NotAdjacent
            option.position = QStyleOptionTab.TabPosition.OnlyOneTab

            option.rect.adjust(0, 0, 0, -overlap)

            # Resetting the ClipRect ensures the overlap isn't drawn.
            painter.setClipRect(QtCore.QRect())

            element = QtWidgets.QStyle.ControlElement.CE_TabBarTab
            style.drawControl(element, option, painter, self)

    def setMaximumHeight(self, maxh: int) -> None:
        self._maximum_height = maxh
        super().setMaximumHeight(maxh)

    def setMaximumSize(self, size: QtCore.QSize) -> None:
        self._maximum_height = size.height()
        super().setMaximumSize(size)

    def box_style(self) -> Style:
        return self._style

    def checkable(self) -> bool:
        return self._checkable

    def checked(self) -> bool:
        return self._checkable and self.checkbox.isChecked()

    def collapsible(self) -> bool:
        return self._collapsible

    def collapsed(self) -> bool:
        return self._collapsible and self._collapsed

    def title(self) -> str:
        return self.title_label.text()

    def set_actions(self, actions: Sequence[QtGui.QAction]) -> None:
        for action in self.actions():
            self.removeAction(action)
        self.addActions(actions)

    def set_checkable(self, checkable: bool) -> None:
        self._checkable = checkable
        self.checkbox.setVisible(checkable)
        if checkable:
            self.frame.setEnabled(self.checkbox.isChecked())
        else:
            self.checkbox.setChecked(False)

    def set_checked(self, checked: bool) -> None:
        if self.checkable():
            self.checkbox.setChecked(checked)

    def set_collapsible(self, collapsible: bool) -> None:
        self._collapsible = collapsible

        if collapsible:
            margins = QtCore.QMargins(4, 4, 4, 4)
            self.set_collapsed(self._collapsed)
        else:
            margins = self._default_margins
            self._expand_more_label.setVisible(False)
            self._expand_less_label.setVisible(False)
        self.header.layout().setContentsMargins(margins)

    def set_collapsed(self, collapsed: bool) -> None:
        if not self.collapsible():
            return

        self._collapsed = collapsed
        if collapsed:
            self.frame.setMaximumHeight(0)
            self._expand_more_label.setVisible(True)
            self._expand_less_label.setVisible(False)
        else:
            self.frame.setMaximumHeight(self._maximum_height)
            self._expand_more_label.setVisible(False)
            self._expand_less_label.setVisible(True)

    def set_box_style(self, style: Style) -> None:
        self._style = style

        if style == CollapsibleBox.Style.SIMPLE:
            self.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        else:
            self.setFrameShape(QtWidgets.QFrame.Shape.NoFrame)

        if style == CollapsibleBox.Style.BUTTON:
            self._layout.setContentsMargins(2, 2, 2, 2)
        else:
            self._layout.setContentsMargins(0, 0, 0, 0)

    def set_title(self, title: str) -> None:
        self.title_label.setText(title)

    def _show_menu(self) -> None:
        relative_pos = self.menu_button.rect().topRight()
        relative_pos.setX(relative_pos.x() + 2)
        position = self.menu_button.mapToGlobal(relative_pos)

        menu = QtWidgets.QMenu(self)
        menu.addActions(self.actions())
        menu.exec_(position)

        self.menu_button.setDown(False)



from __future__ import annotations

import logging
from collections.abc import Sequence
from functools import partial
from typing import Callable

from qt_material_icons import MaterialIcon
from qtpy import QtCore, QtGui, QtWidgets

from .editor import ParameterForm
from .widgets import ParameterWidget, StringParameter

logger = logging.getLogger(__name__)


class DragItem(QtWidgets.QFrame):
    drag_started = QtCore.Signal()
    drag_ended = QtCore.Signal()
    removed = QtCore.Signal()

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent=parent)

        self._widget = None

        self._init_ui()

    def _init_ui(self) -> None:
        # Force the TabFocus FocusPolicy to allow setting TabOrder.
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.TabFocus)
        self.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        self.setAutoFillBackground(True)

        self._layout = QtWidgets.QHBoxLayout()
        self.setLayout(self._layout)

        self.drag_label = QtWidgets.QLabel()
        pixmap = MaterialIcon('drag_handle').pixmap()
        self.drag_label.setPixmap(pixmap)
        self._layout.addWidget(self.drag_label)
        self._layout.setStretch(0, 0)

        self._widget = QtWidgets.QWidget()
        self._layout.addWidget(self._widget)
        self._layout.setStretch(1, 1)

        remove_button = QtWidgets.QToolButton()
        remove_button.setIcon(MaterialIcon('delete'))
        remove_button.clicked.connect(self.removed)
        remove_button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self._layout.addWidget(remove_button)

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        if self.childAt(event.pos()) == self.drag_label:
            self.drag_started.emit()
            event.accept()
            return
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        self.drag_ended.emit()
        super().mouseReleaseEvent(event)

    def widget(self) -> QtWidgets.QWidget:
        return self._widget

    def set_widget(self, widget: QtWidgets.QWidget) -> None:
        if self._widget:
            self._layout.removeWidget(self._widget)
            self._widget.deleteLater()
        self._widget = widget
        self._layout.insertWidget(1, self._widget)
        self._layout.setStretch(1, 1)
        self.setFocusProxy(self._widget)


class DragWidget(QtWidgets.QWidget):
    item_added = QtCore.Signal(DragItem)
    item_removed = QtCore.Signal()
    item_moved = QtCore.Signal(DragItem)

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._factory = QtWidgets.QWidget
        self._placeholder = None
        self._drag_item = None
        self._drag_offset = 0
        self._drag_start_index = -1

        self._init_ui()

    def _init_ui(self) -> None:
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(4)
        self.setLayout(layout)

        self._item_layout = QtWidgets.QVBoxLayout()
        layout.addLayout(self._item_layout)

        add_frame = QtWidgets.QFrame()
        layout.addWidget(add_frame)

        add_layout = QtWidgets.QHBoxLayout()
        add_frame.setLayout(add_layout)

        add_layout.addStretch()

        add_button = QtWidgets.QToolButton()
        add_button.setIcon(MaterialIcon('add'))
        add_button.clicked.connect(self.add_item)
        add_button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        add_layout.addWidget(add_button)

        layout.addStretch()

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        if self._drag_item:
            self._drag_move(event.pos())

    def factory(self) -> type[QtWidgets.QWidget] | Callable[[], QtWidgets.QWidget]:
        return self._factory

    def set_factory(
        self, factory: type[QtWidgets.QWidget] | Callable[[], QtWidgets.QWidget]
    ) -> None:
        self._factory = factory

    def add_item(self) -> DragItem:
        """Add a new item to the layout."""

        item = DragItem()
        item.drag_started.connect(partial(self._drag_started, item))
        item.drag_ended.connect(self._drag_ended)
        item.removed.connect(partial(self._remove_item, item))

        widget = self._factory()
        item.set_widget(widget)

        self._item_layout.addWidget(item)
        self.item_added.emit(item)
        return item

    def items(self) -> tuple[DragItem, ...]:
        """Return all DragItems in the layout."""

        items = []
        for i in range(self._item_layout.count()):
            if item := self._item_layout.itemAt(i):
                widget = item.widget()
                if isinstance(widget, DragItem):
                    items.append(widget)
        return tuple(items)

    def clear(self) -> None:
        """Clear the layout."""

        while item := self._item_layout.takeAt(0):
            if widget := item.widget():
                widget.deleteLater()

    def _remove_item(self, item: DragItem) -> None:
        self._item_layout.removeWidget(item)
        item.deleteLater()
        self.item_removed.emit()

    def _drag_started(self, item: DragItem) -> None:
        """
        Handle the start of a drag event. Store the current offset and create a
        placeholder.
        """

        self.setCursor(QtGui.QCursor(QtCore.Qt.CursorShape.SizeAllCursor))

        self._drag_item = item
        local_position = self.mapFromGlobal(QtGui.QCursor.pos())
        self._drag_offset = item.y() - local_position.y()

        # Create placeholder
        self._placeholder = QtWidgets.QWidget()
        self._placeholder.setMinimumHeight(item.height())
        index = self._item_layout.indexOf(item)
        self._drag_start_index = index
        if index >= 0:
            self._item_layout.insertWidget(index, self._placeholder)

        # Float the DragItem
        self._item_layout.removeWidget(item)
        item.raise_()

    def _drag_move(self, position: QtCore.QPoint) -> None:
        """
        Handle the move of the drag event. Move the DragItem with the mouse and keep
        the placeholder underneath the mouse.
        """

        pos_y = position.y() + self._drag_offset
        self._drag_item.move(self._drag_item.x(), pos_y)

        index = self._get_index(position)
        if index >= 0:
            self._move_item(self._placeholder, index)

    def _drag_ended(self) -> None:
        """
        Handle the end of a drag event. Remove the placeholder and re-insert the
        DragItem.
        """

        if not self._drag_item:
            # This was not a drag event.
            return

        index = -1
        if self._placeholder:
            index = self._item_layout.indexOf(self._placeholder)
            self._item_layout.removeWidget(self._placeholder)
            self._placeholder.deleteLater()
            self._placeholder = None

        if index < 0:
            index = self._drag_start_index
        # Ensure the DragItem is always inserted back into the layout.
        index = max(0, index)
        self._item_layout.insertWidget(index, self._drag_item)
        self._drag_item = None

        self._refresh_tab_order()
        self.unsetCursor()

        if index != self._drag_start_index:
            self.item_moved.emit(self._drag_item)
        self._drag_start_index = -1

    def _get_index(self, position: QtCore.QPoint) -> int:
        """Return the index in the layout for a position."""

        i = -1
        for i in range(self._item_layout.count()):
            if widget := self._item_layout.itemAt(i).widget():
                if position.y() < widget.y() + widget.height():
                    return i
        return i

    def _move_item(self, widget: QtWidgets.QWidget, index: int) -> None:
        """Move the widget to the given index, but only if needed."""

        current_index = self._item_layout.indexOf(widget)
        if current_index != index:
            self._item_layout.insertWidget(index, widget)

    def _refresh_tab_order(self) -> None:
        """Refresh the TabOrder from the start to the end of the layout."""

        first = None
        for i in range(self._item_layout.count()):
            if item := self._item_layout.itemAt(i):
                if widget := item.widget():
                    if first:
                        self.setTabOrder(first, widget)
                    first = widget


class ListParameter(ParameterWidget):
    _value: tuple = ()
    _default: tuple = ()
    _factory: type[ParameterWidget | ParameterForm] | Callable | None = None

    def _init_ui(self) -> None:
        self.drag = DragWidget()
        self.drag.item_added.connect(self._item_added)
        self.drag.item_removed.connect(self._items_changed)
        self.drag.item_moved.connect(lambda: self._items_changed())
        self.setFocusProxy(self.drag)
        self._layout.addWidget(self.drag)

    def factory(self) -> type[ParameterWidget | ParameterForm] | Callable | None:
        return self._factory

    def set_factory(
        self, factory: type[ParameterWidget | ParameterForm] | Callable
    ) -> None:
        self._factory = factory
        self.drag.set_factory(factory)
        self.drag.clear()

    def value(self) -> tuple:
        return super().value()

    def set_value(self, value: Sequence) -> None:
        self.drag.clear()
        for v in value:
            item = self.drag.add_item()
            widget = item.widget()
            widget.blockSignals(True)
            if isinstance(widget, ParameterWidget):
                widget.set_value(v)
            elif isinstance(widget, ParameterForm):
                widget.set_values(v)
            widget.blockSignals(False)

        if not isinstance(value, tuple):
            value = tuple(value)
        super().set_value(value)

    def _item_added(self, item: DragItem) -> None:
        widget = item.widget()
        if isinstance(widget, ParameterWidget):
            widget.value_changed.connect(self._items_changed)
        elif isinstance(widget, ParameterForm):
            widget.parameter_changed.connect(self._items_changed)

    def _items_changed(self) -> None:
        values = self._get_values()
        super().set_value(values)

    def _get_values(self) -> tuple:
        """Return the values of all items in the DragWidget as a tuple."""

        values = []
        for item in self.drag.items():
            widget = item.widget()
            if isinstance(widget, ParameterWidget):
                values.append(widget.value())
            elif isinstance(widget, ParameterForm):
                values.append(widget.values())
        return tuple(values)


class StringListParameter(StringParameter):
    _value: tuple[str, ...] = ()
    _default: tuple[str, ...] = ()
    _area: bool = True

    def value(self) -> tuple[str, ...]:
        return super().value()

    def set_value(self, value: Sequence[str]) -> None:
        if not isinstance(value, tuple):
            value = tuple(value)
        ParameterWidget.set_value(self, value)

        self.text.blockSignals(True)
        if isinstance(self.text, QtWidgets.QPlainTextEdit):
            text = '\n'.join(value)
            self.text.setPlainText(text)
            self._refresh_height()
        elif isinstance(self.text, QtWidgets.QLineEdit):
            text = ' '.join(value)
            self.text.setText(text)
        self.text.blockSignals(False)

    def _action_triggered(self, action: QtGui.QAction) -> None:
        data = action.data()
        value = str(data)
        if self._menu_mode == StringParameter.MenuMode.REPLACE:
            self.set_value((value,))
        elif self._menu_mode == StringParameter.MenuMode.TOGGLE:
            values = self._value
            if value in values:
                values = (v for v in values if v != value)
            else:
                values = (*values, value)
            self.set_value(values)

    def _editing_finished(self) -> None:
        if isinstance(self.text, QtWidgets.QPlainTextEdit):
            values = self.text.toPlainText().split('\n')
        elif isinstance(self.text, QtWidgets.QLineEdit):
            values = self.text.text().split(' ')
        else:
            return

        values = tuple(v for v in values if v)
        ParameterWidget.set_value(self, values)



from __future__ import annotations

from collections.abc import Collection, Mapping, Sequence
from enum import Enum, EnumMeta, auto
from functools import partial
from typing import Any, Callable

from qt_material_icons import MaterialIcon
from qtpy import QtCore, QtGui, QtWidgets

from . import utils
from .inputs import (
    FloatLineEdit,
    FloatSlider,
    IntLineEdit,
    IntSlider,
    RatioButton,
    TextEdit,
)
from .resizegrip import ResizeGrip

MIN_SLIDER_WIDTH = 200


class ParameterWidget(QtWidgets.QWidget):
    value_changed = QtCore.Signal(object)

    _value: Any = None
    _default: Any = None
    _name: str = ''
    _label: str = ''
    _tooltip: str = ''

    def __init__(self, name: str = '', parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._init_layout()
        self._init_ui()

        if name:
            self.set_name(name)
            self.set_label(utils.title(name))

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._name!r})'

    def _init_layout(self) -> None:
        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(QtCore.QMargins())
        self.setLayout(self._layout)

    def _init_ui(self) -> None: ...

    def value(self) -> Any:
        return self._value

    def set_value(self, value: Any) -> None:
        if value != self._value:
            self._value = value
            self.value_changed.emit(value)

    def default(self) -> Any:
        return self._default

    def set_default(self, default: Any) -> None:
        self.set_value(default)
        self._default = self.value()

    def label(self) -> str:
        return self._label

    def set_label(self, label: str) -> None:
        self._label = label

    def name(self) -> str:
        return self._name

    def set_name(self, name: str) -> None:
        self._name = name

    def tooltip(self) -> str:
        return self._tooltip

    def set_tooltip(self, tooltip: str) -> None:
        self._tooltip = tooltip

    def reset(self) -> None:
        self.set_value(self.default())


class IntParameter(ParameterWidget):
    value_changed = QtCore.Signal(int)

    _value: int = 0
    _default: int = 0
    _slider_min: int = 0
    _slider_max: int = 10
    _line_min: int | None = None
    _line_max: int | None = None
    _slider_visible: bool = True
    _commit_on_edit: bool = False
    _step_factor: int = 2

    def _init_ui(self) -> None:
        # Line
        self.line = IntLineEdit(self)
        self.line.set_value(self._value)
        self.line.value_changed.connect(self._line_value_changed)
        self._layout.addWidget(self.line)

        # Slider
        self.slider = IntSlider()
        self.slider.set_maximum(self._slider_max)
        self.slider.value_changed.connect(self._slider_value_changed)
        # Prevent any size changes when slider shows
        self.slider.setMaximumHeight(self.line.minimumSizeHint().height())
        self._layout.addWidget(self.slider)
        self._layout.setStretch(1, 1)

        self.setFocusProxy(self.line)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        super().resizeEvent(event)
        self._toggle_slider(True)

    def commit_on_edit(self) -> bool:
        return self._commit_on_edit

    def set_commit_on_edit(self, commit_on_edit: bool) -> None:
        self._commit_on_edit = commit_on_edit
        self.line.commit_on_edit = commit_on_edit

    def line_min(self) -> int | None:
        return self._line_min

    def set_line_min(self, line_min: int | None) -> None:
        self._line_min = line_min
        self.line.set_minimum(line_min)

    def line_max(self) -> int | None:
        return self._line_max

    def set_line_max(self, line_max: int | None) -> None:
        self._line_max = line_max
        self.line.set_maximum(line_max)

    def slider_min(self) -> int:
        return self._slider_min

    def set_slider_min(self, slider_min: int) -> None:
        self._slider_min = slider_min
        self.slider.set_minimum(slider_min)

    def slider_max(self) -> int:
        return self._slider_max

    def set_slider_max(self, slider_max: int) -> None:
        self._slider_max = slider_max
        self.slider.set_maximum(slider_max)

    def slider_visible(self) -> bool:
        return self._slider_visible

    def set_slider_visible(self, slider_visible: bool) -> None:
        self._slider_visible = slider_visible
        self._toggle_slider(slider_visible)

    def step_factor(self) -> int:
        return self._step_factor

    def set_step_factor(self, factor: int) -> None:
        self._step_factor = factor
        self.slider.set_step_factor(factor)

    def value(self) -> int:
        return super().value()

    def set_value(self, value: int) -> None:
        super().set_value(value)
        self._set_line_value(value)
        self._set_slider_value(value)

    def _line_value_changed(self, value: int) -> None:
        super().set_value(value)
        self._set_slider_value(value)

    def _slider_value_changed(self, value: int) -> None:
        super().set_value(value)
        self._set_line_value(value)

    def _set_line_value(self, value: int) -> None:
        self.line.blockSignals(True)
        self.line.set_value(value)
        self.line.blockSignals(False)

    def _set_slider_value(self, value: int) -> None:
        self.slider.blockSignals(True)
        self.slider.set_value(value)
        self.slider.blockSignals(False)

    def _toggle_slider(self, value: bool) -> None:
        has_space = self.size().width() > MIN_SLIDER_WIDTH
        self.slider.setVisible(self._slider_visible and value and has_space)


class FloatParameter(IntParameter):
    value_changed = QtCore.Signal(float)

    _value: float = 0
    _default: float = 0
    _slider_min: float = 0
    _slider_max: float = 1
    _line_min: float | None = None
    _line_max: float | None = None
    _decimals: int = 4

    def _init_ui(self) -> None:
        # line
        self.line = FloatLineEdit(self)
        self.line.value_changed.connect(self._line_value_changed)
        self.line.set_value(self._value)
        self.line.set_decimals(self._decimals)
        self._layout.addWidget(self.line)

        # slider
        self.slider = FloatSlider()
        self.slider.set_maximum(self._slider_max)
        self.slider.value_changed.connect(self._slider_value_changed)
        # prevent any size changes when slider shows
        self.slider.setMaximumHeight(self.line.minimumSizeHint().height())
        self._layout.addWidget(self.slider)
        self._layout.setStretch(1, 1)

        self.setFocusProxy(self.line)

    def decimals(self) -> int:
        return self._decimals

    def set_decimals(self, decimals: int) -> None:
        self._decimals = decimals
        self.line.set_decimals(decimals)

    def line_min(self) -> float:
        return super().line_min()

    def set_line_min(self, line_min: float | None) -> None:
        super().set_line_min(line_min)

    def line_max(self) -> float:
        return super().line_max()

    def set_line_max(self, line_max: float | None) -> None:
        super().set_line_max(line_max)

    def slider_min(self) -> float:
        return super().slider_min()

    def set_slider_min(self, slider_min: float) -> None:
        super().set_slider_min(slider_min)  # noqa

    def slider_max(self) -> float:
        return super().slider_max()

    def set_slider_max(self, slider_max: float) -> None:
        super().set_slider_max(slider_max)  # noqa

    def value(self) -> float:
        return super().value()

    def set_value(self, value: float) -> None:
        super().set_value(value)  # noqa


class StringParameter(ParameterWidget):
    class MenuMode(Enum):
        REPLACE = auto()
        TOGGLE = auto()

    value_changed = QtCore.Signal(str)

    _value: str = ''
    _default: str = ''
    _placeholder: str = ''
    _area: bool = False
    _menu: QtWidgets.QMenu | None = None
    _menu_data: Collection | None = None
    _menu_mode: MenuMode = MenuMode.REPLACE

    def _init_ui(self) -> None:
        self._init_text()

        self.menu_button = QtWidgets.QToolButton()
        self.menu_button.setAutoRaise(True)
        self._layout.addWidget(self.menu_button)
        self.menu_button.hide()

    def _init_text(self) -> None:
        if self._area:
            self.text = TextEdit()
            self.text.editing_finished.connect(self._editing_finished)
            resize_grip = ResizeGrip(self.text)
            # Initialize the ResizeGrip to allow resizing smaller
            _ = resize_grip.min_size
        else:
            self.text = QtWidgets.QLineEdit()
            self.text.editingFinished.connect(self._editing_finished)
        self._layout.insertWidget(0, self.text)
        self.setFocusProxy(self.text)

    def area(self) -> bool:
        return self._area

    def set_area(self, area: bool) -> None:
        if area != self._area:
            self._area = area
            self._layout.removeWidget(self.text)
            self.text.deleteLater()
            self._init_text()

    def menu(self) -> QtWidgets.QMenu:
        if self._menu is None:
            self._menu = self._build_menu(self._menu_data)
        return self._menu

    def set_menu(self, menu: Collection | None) -> None:
        """
        Set the menu used to populate the parameter.
        The menu can be a Sequence of values or a dictionary of label: value pairs.
        """

        self._menu_data = menu
        self._menu = None

        # Update menu
        if not self._area and self._menu_data is not None:
            if not self.menu_button.defaultAction():
                # build dynamically for optimization
                icon = MaterialIcon('expand_more')
                action = QtGui.QAction(icon, 'Fill', self)
                action.triggered.connect(self._show_menu)
                self.menu_button.setDefaultAction(action)
            self.menu_button.show()
        else:
            self.menu_button.hide()

    def menu_mode(self) -> MenuMode:
        return self._menu_mode

    def set_menu_mode(self, mode: MenuMode) -> None:
        self._menu_mode = mode

    def placeholder(self) -> str:
        return self._placeholder

    def set_placeholder(self, placeholder: str) -> None:
        self._placeholder = placeholder
        self.text.setPlaceholderText(placeholder)

    def value(self) -> str:
        return super().value()

    def set_value(self, value: str) -> None:
        super().set_value(value)
        self.text.blockSignals(True)
        if isinstance(self.text, QtWidgets.QPlainTextEdit):
            self.text.setPlainText(value)
            self._refresh_height()
        elif isinstance(self.text, QtWidgets.QLineEdit):
            self.text.setText(value)
        self.text.blockSignals(False)

    def _action_triggered(self, action: QtGui.QAction) -> None:
        data = action.data()
        value = str(data)
        if self._menu_mode == StringParameter.MenuMode.REPLACE:
            self.set_value(value)
        elif self._menu_mode == StringParameter.MenuMode.TOGGLE:
            values = self._value.split(' ')
            if value in values:
                values = (v for v in values if v != value)
            else:
                values.append(value)
            self.set_value(' '.join(values))

    def _build_menu(
        self, content: Collection, menu: QtWidgets.QMenu | None = None
    ) -> QtWidgets.QMenu:
        """Recursively build the QMenu from content."""

        if menu is None:
            menu = QtWidgets.QMenu(self)
        if isinstance(content, Sequence):
            content = {i: i for i in content}
        for label, data in content.items():
            if isinstance(data, Mapping):
                sub_menu = menu.addMenu(label)
                self._build_menu(data, sub_menu)
            else:
                action = QtGui.QAction(label, self)
                action.setData(data)
                action.triggered.connect(partial(self._action_triggered, action))
                menu.addAction(action)
        return menu

    def _editing_finished(self) -> None:
        if isinstance(self.text, QtWidgets.QPlainTextEdit):
            super().set_value(self.text.toPlainText())
        elif isinstance(self.text, QtWidgets.QLineEdit):
            super().set_value(self.text.text())

    def _refresh_height(self) -> None:
        if isinstance(self.text, QtWidgets.QPlainTextEdit):
            line_count = self.text.document().lineCount() + 1
            metrics = self.text.fontMetrics()
            line_spacing = metrics.lineSpacing()
            height = (
                line_count * line_spacing
                + self.text.contentsMargins().top()
                + self.text.contentsMargins().bottom()
            )
            height = max(height, self.text.minimumHeight())
            self.text.setFixedHeight(height)

    def _show_menu(self) -> None:
        relative_pos = self.menu_button.rect().topRight()
        relative_pos.setX(relative_pos.x() + 2)
        position = self.menu_button.mapToGlobal(relative_pos)

        menu = self.menu()
        menu.exec_(position)
        self.menu_button.setDown(False)


class PathParameter(ParameterWidget):
    class Method(Enum):
        OPEN_FILE = auto()
        SAVE_FILE = auto()
        EXISTING_DIR = auto()

    OPEN_FILE = Method.OPEN_FILE
    SAVE_FILE = Method.SAVE_FILE
    EXISTING_DIR = Method.EXISTING_DIR

    value_changed = QtCore.Signal(str)

    _value: str = ''
    _default: str = ''
    _dir_fallback: str = ''
    _method: Method = Method.OPEN_FILE

    def _init_ui(self) -> None:
        self.line = QtWidgets.QLineEdit()
        self.line.editingFinished.connect(self._editing_finished)
        self._layout.addWidget(self.line)

        self.button = QtWidgets.QToolButton()
        self.button.setIcon(MaterialIcon('file_open'))
        self.button.clicked.connect(self.browse)
        self._layout.addWidget(self.button)

        self._layout.setStretch(0, 1)
        self.setFocusProxy(self.line)

    def browse(self) -> None:
        start_dir = self._value or self._dir_fallback
        if self._method == PathParameter.Method.OPEN_FILE:
            path, filters = QtWidgets.QFileDialog.getOpenFileName(
                parent=self, caption='Open File', dir=start_dir
            )
        elif self._method == PathParameter.Method.SAVE_FILE:
            path, filters = QtWidgets.QFileDialog.getSaveFileName(
                parent=self, caption='Save File', dir=start_dir, filter='*.*'
            )
        elif self._method == PathParameter.Method.EXISTING_DIR:
            path = QtWidgets.QFileDialog.getExistingDirectory(
                parent=self, caption='Select Directory', dir=start_dir
            )
        else:
            return

        if path:
            self.set_value(path)

    def dir_fallback(self) -> str:
        return self._dir_fallback

    def set_dir_fallback(self, dir_fallback: str) -> None:
        self._dir_fallback = dir_fallback

    def method(self) -> Method:
        return self._method

    def set_method(self, method: Method) -> None:
        self._method = method

    def value(self) -> str:
        return super().value()

    def set_value(self, value: str) -> None:
        super().set_value(value)
        self.line.blockSignals(True)
        self.line.setText(value)
        self.line.blockSignals(False)

    def _editing_finished(self) -> None:
        value = self.line.text()
        super().set_value(value)


class ComboParameter(ParameterWidget):
    _value: Any = None
    _default: Any = None
    _items: tuple = ()

    def _init_ui(self) -> None:
        self.combo = QtWidgets.QComboBox()
        self.combo.currentIndexChanged.connect(self._current_index_changed)
        self.combo.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )

        self._layout.addWidget(self.combo)
        self.setFocusProxy(self.combo)

    def items(self) -> tuple:
        return self._items

    def set_items(self, items: Collection) -> None:
        if isinstance(items, Mapping):
            items = tuple(items.items())
        else:
            items = tuple(i if isinstance(i, tuple) else (i, i) for i in items)

        self._items = items
        self._refresh_items()
        try:
            default = items[0][1]
        except (IndexError, TypeError):
            default = None
        self.set_default(default)
        self.set_value(default)

    def value(self) -> Any:
        return super().value()

    def set_value(self, value: Any) -> None:
        index = self._index_from_value(value)
        value = self.combo.itemData(index)
        super().set_value(value)
        self.combo.blockSignals(True)
        self.combo.setCurrentIndex(index)
        self.combo.blockSignals(False)

    def _current_index_changed(self, index: int) -> None:
        value = self.combo.itemData(index)
        super().set_value(value)

    def _index_from_value(self, value: Any) -> int:
        """Return the index for a value, searching text and data."""
        if value is None:
            return -1

        if isinstance(value, str):
            index = self.combo.findText(value)
        else:
            index = -1

        if index < 0:
            index = self.combo.findData(value)
        return index

    def _refresh_items(self) -> None:
        self.combo.blockSignals(True)
        for index in reversed(range(self.combo.count())):
            self.combo.removeItem(index)
        for label, data in self._items:
            self.combo.addItem(label, data)
        self.combo.blockSignals(False)


class EnumParameter(ParameterWidget):
    _value: Enum | None = None
    _default: Enum | None = None
    _formatter: Callable | None = None
    _enum: EnumMeta | None = None

    def __init__(self, name: str = '', parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(name=name, parent=parent)

        self._formatter = lambda member: utils.title(member.name)

    def _init_ui(self) -> None:
        self.combo = QtWidgets.QComboBox()
        self.combo.currentIndexChanged.connect(self._current_index_changed)
        self.combo.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )

        self._layout.addWidget(self.combo)
        self.setFocusProxy(self.combo)

    def enum(self) -> EnumMeta:
        return self._enum

    def set_enum(self, enum: EnumMeta) -> None:
        self._enum = enum
        self._update_items()
        if self._enum:
            default = tuple(self._enum)[0]
        else:
            default = None
        self.set_default(default)
        self.set_value(default)

    def formatter(self) -> Callable:
        return self._formatter

    def set_formatter(self, formatter: Callable) -> None:
        self._formatter = formatter
        index = self.combo.currentIndex()
        self._update_items()
        self.combo.setCurrentIndex(index)

    def value(self) -> Enum | None:
        return super().value()

    def set_value(self, value: Any) -> None:
        value = self._enum_from_value(value)
        super().set_value(value)

        self.combo.blockSignals(True)
        if value is None:
            index = -1
        else:
            index = self.combo.findData(value.value)
        self.combo.setCurrentIndex(index)
        self.combo.blockSignals(False)

    def _current_index_changed(self, index: int) -> None:
        value = self.combo.itemData(index)
        value = self._enum_from_value(value)
        super().set_value(value)

    def _enum_from_value(self, value: Any) -> Enum | None:
        try:
            # value is Enum
            if isinstance(value, self._enum):
                return value
        except TypeError:
            pass

        try:
            # value is Enum.name
            return self._enum[value]
        except KeyError:
            pass
        except TypeError:
            return None

        try:
            # value is Enum.value
            return self._enum(value)
        except (ValueError, TypeError):
            return None

    def _update_items(self) -> None:
        self.combo.blockSignals(True)
        for index in reversed(range(self.combo.count())):
            self.combo.removeItem(index)

        if isinstance(self._enum, EnumMeta):
            for member in self._enum:
                label = self._formatter(member)
                self.combo.addItem(label, member.value)
        self.combo.blockSignals(False)


class BoolParameter(ParameterWidget):
    value_changed = QtCore.Signal(bool)

    _value: bool = False
    _default: bool = False

    def _init_ui(self) -> None:
        self.checkbox = QtWidgets.QCheckBox()
        self.checkbox.toggled.connect(super().set_value)
        self._layout.addWidget(self.checkbox)
        self._layout.addStretch()
        self.setFocusProxy(self.checkbox)

    def value(self) -> bool:
        return super().value()

    def set_value(self, value: bool) -> None:
        super().set_value(value)
        self.checkbox.blockSignals(True)
        self.checkbox.setChecked(value)
        self.checkbox.blockSignals(False)


class MultiIntParameter(IntParameter):
    value_changed = QtCore.Signal(tuple)

    _count: int = 2
    _value: tuple[int, ...] = (0, 0)
    _default: tuple[int, ...] = (0, 0)
    _keep_ratio: bool = True
    _ratio_visible: bool = True

    def _init_ui(self) -> None:
        # Lines
        self.lines = []
        for i in range(self._count):
            line = IntLineEdit()
            line.set_value(0)
            line.value_changed.connect(self._line_value_changed)
            self._layout.addWidget(line)
            self.lines.append(line)

        # Slider
        self.slider = IntSlider()
        self.slider.set_maximum(self._slider_max)
        self.slider.value_changed.connect(self._slider_value_changed)
        # Prevent any size changes when slider shows
        line_height = self.lines[0].minimumSizeHint().height()
        self.slider.setMaximumHeight(line_height)
        self._layout.addWidget(self.slider)
        self._layout.setStretch(self._count, 1)

        # Keep ratio button
        self.keep_ratio_button = RatioButton()
        self.keep_ratio_button.setMaximumSize(line_height, line_height)
        self.keep_ratio_button.toggled.connect(self.set_keep_ratio)
        self._layout.addWidget(self.keep_ratio_button)

        self.setFocusProxy(self.lines[0])
        self.set_keep_ratio(self._keep_ratio)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        QtWidgets.QWidget.resizeEvent(self, event)
        if self._keep_ratio:
            self._toggle_slider(True)

    def set_commit_on_edit(self, commit_on_edit: bool) -> None:
        self._commit_on_edit = commit_on_edit
        for line in self.lines:
            line.commit_on_edit = commit_on_edit

    def set_keep_ratio(self, keep_ratio: bool) -> None:
        self._keep_ratio = keep_ratio
        self.keep_ratio_button.setChecked(keep_ratio)
        for line in self.lines[1:]:
            line.setVisible(not keep_ratio)
            if line.value() != self.lines[0].value():
                line.set_value(self.lines[0].value())
        self._toggle_slider(keep_ratio)

    def set_line_min(self, line_min: int) -> None:
        self._line_min = line_min
        for line in self.lines:
            line.set_minimum(line_min)

    def set_line_max(self, line_max: int) -> None:
        self._line_max = line_max
        for line in self.lines:
            line.set_maximum(line_max)

    def set_ratio_visible(self, ratio_visible: bool) -> None:
        self._ratio_visible = ratio_visible
        self.keep_ratio_button.setVisible(ratio_visible)
        for line in self.lines[1:]:
            line.setVisible(not ratio_visible)
        if not ratio_visible:
            self.set_keep_ratio(False)

    def value(self) -> tuple[int, ...]:
        return ParameterWidget.value(self)

    def set_value(self, value: Sequence) -> None:
        if isinstance(value, Sequence):
            values = value
        else:
            values = self._cast_to_tuple(value)
        if isinstance(self, ColorParameter):
            pass
        if not all(values[0] == x for x in values):
            self.set_keep_ratio(False)
        if self._keep_ratio:
            values = (values[0],) * self._count
        ParameterWidget.set_value(self, self._cast_to_type(values))
        self._set_slider_value(values[0])
        self._set_line_values(values)

    def _line_value_changed(self, value: int) -> None:
        if self._keep_ratio:
            values = (self.lines[0].value(),) * self._count
            for line in self.lines[1:]:
                line.set_value(values[0])
        else:
            values = tuple(line.value() for line in self.lines)

        value = self._cast_to_type(values)
        ParameterWidget.set_value(self, value)
        self._set_slider_value(values[0])

    def _slider_value_changed(self, value: int) -> None:
        values = (value,) * self._count
        value = self._cast_to_type(values)
        ParameterWidget.set_value(self, value)
        self._set_line_values(values)

    def _set_line_values(self, values: tuple[int, ...]) -> None:
        for line, value in zip(self.lines, values):
            line.blockSignals(True)
            line.set_value(value)
            line.blockSignals(False)

    def _cast_to_tuple(self, values: Any) -> tuple[int, ...]:
        return values

    def _cast_to_type(self, values: tuple[int, ...]) -> Any:
        return values


class MultiFloatParameter(MultiIntParameter):
    _value: tuple[float, ...] = (0, 0)
    _default: tuple[float, ...] = (0, 0)
    _line_min: float | None = None
    _line_max: float | None = None
    _slider_min: float = 0
    _slider_max: float = 1
    _decimals: int = 4

    def _init_ui(self) -> None:
        # Lines
        self.lines = []
        for i in range(self._count):
            line = FloatLineEdit()
            line.set_value(0)
            line.set_decimals(self._decimals)
            line.value_changed.connect(self._line_value_changed)
            self._layout.addWidget(line)
            self.lines.append(line)

        # Slider
        self.slider = FloatSlider()
        self.slider.set_maximum(self._slider_max)
        self.slider.value_changed.connect(self._slider_value_changed)
        # Prevent any size changes when slider shows
        line_height = self.lines[0].minimumSizeHint().height()
        self.slider.setMaximumHeight(line_height)
        self._layout.addWidget(self.slider)
        self._layout.setStretch(self._count, 1)

        # Keep ratio button
        self.keep_ratio_button = RatioButton()
        self.keep_ratio_button.setMaximumSize(line_height, line_height)
        self.keep_ratio_button.toggled.connect(self.set_keep_ratio)
        self._layout.addWidget(self.keep_ratio_button)

        self.setFocusProxy(self.lines[0])
        self.set_keep_ratio(self._keep_ratio)

    def decimals(self) -> int:
        return self._decimals

    def set_decimals(self, decimals: int) -> None:
        self._decimals = decimals
        for line in self.lines:
            line.set_decimals(decimals)

    def line_min(self) -> float:
        return super().line_min()

    def set_line_min(self, line_min: float) -> None:
        super().set_line_min(line_min)  # noqa

    def line_max(self) -> float:
        return super().line_max()

    def set_line_max(self, line_max: float) -> None:
        super().set_line_max(line_max)  # noqa

    def slider_min(self) -> float:
        return super().slider_min()

    def set_slider_min(self, slider_min: float) -> None:
        super().set_slider_min(slider_min)  # noqa

    def slider_max(self) -> float:
        return super().slider_max()

    def set_slider_max(self, slider_max: float) -> None:
        super().set_slider_max(slider_max)  # noqa

    def value(self) -> tuple[float, ...]:
        return super().value()

    def set_value(self, value: tuple[float, ...]) -> None:
        super().set_value(value)


class PointParameter(MultiIntParameter):
    value_changed = QtCore.Signal(QtCore.QPoint)

    _value: QtCore.QPoint = QtCore.QPoint(0, 0)
    _default: QtCore.QPoint = QtCore.QPoint(0, 0)
    _slider_visible: bool = False
    _ratio_visible: bool = False

    def _init_ui(self) -> None:
        super()._init_ui()
        self.set_slider_visible(self._slider_visible)
        self.set_ratio_visible(self._ratio_visible)

    def set_value(self, value: QtCore.QPoint | Sequence) -> None:
        super().set_value(value)

    def value(self) -> QtCore.QPoint:
        return super().value()  # noqa

    def _cast_to_type(self, values: tuple[int, ...]) -> QtCore.QPoint:
        return QtCore.QPoint(*values[:2])

    def _cast_to_tuple(self, value: QtCore.QPoint) -> tuple[int, ...]:
        return value.x(), value.y()


class PointFParameter(MultiFloatParameter):
    value_changed = QtCore.Signal(QtCore.QPointF)

    _value: QtCore.QPointF = QtCore.QPointF(0, 0)
    _default: QtCore.QPointF = QtCore.QPointF(0, 0)
    _slider_visible: bool = False
    _ratio_visible: bool = False

    def _init_ui(self) -> None:
        super()._init_ui()
        self.set_slider_visible(self._slider_visible)
        self.set_ratio_visible(self._ratio_visible)

    def set_value(self, value: QtCore.QPointF | Sequence) -> None:
        super().set_value(value)  # noqa

    def value(self) -> QtCore.QPointF:
        return super().value()  # noqa

    def _cast_to_type(self, values: tuple[float, ...]) -> QtCore.QPointF:
        return QtCore.QPointF(*values[:2])

    def _cast_to_tuple(self, value: QtCore.QPointF) -> tuple[float, ...]:
        return value.x(), value.y()


class SizeParameter(MultiIntParameter):
    value_changed = QtCore.Signal(QtCore.QSize)

    _value: QtCore.QSize = QtCore.QSize(0, 0)
    _default: QtCore.QSize = QtCore.QSize(0, 0)

    def set_value(self, value: QtCore.QSize | Sequence) -> None:
        super().set_value(value)

    def value(self) -> QtCore.QSize:
        return super().value()  # noqa

    def _cast_to_type(self, values: tuple[int, ...]) -> QtCore.QSize:
        return QtCore.QSize(*values[:2])

    def _cast_to_tuple(self, value: QtCore.QSize) -> tuple[int, ...]:
        return value.width(), value.height()


class SizeFParameter(MultiFloatParameter):
    value_changed = QtCore.Signal(QtCore.QSizeF)

    _value: QtCore.QSizeF = QtCore.QSizeF(0, 0)
    _default: QtCore.QSizeF = QtCore.QSizeF(0, 0)

    def set_value(self, value: QtCore.QSizeF | Sequence) -> None:
        super().set_value(value)  # noqa

    def value(self) -> QtCore.QSizeF:
        return super().value()  # noqa

    def _cast_to_type(self, values: tuple[float, ...]) -> QtCore.QSizeF:
        return QtCore.QSizeF(*values[:2])

    def _cast_to_tuple(self, value: QtCore.QSizeF) -> tuple[float, ...]:
        return value.width(), value.height()


class ColorParameter(MultiFloatParameter):
    value_changed = QtCore.Signal(QtGui.QColor)

    _count: int = 3
    _value: QtGui.QColor = QtGui.QColor(0, 0, 0)
    _default: QtGui.QColor = QtGui.QColor(0, 0, 0)
    _color_min: float | None = 0
    _color_max: float | None = 1
    _decimals: int = 2

    def _init_ui(self) -> None:
        super()._init_ui()

        for line in self.lines:
            line.set_maximum(self._color_max)

        self.button = QtWidgets.QPushButton()
        self.button.clicked.connect(self.select_color)
        self.button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        size = self.button.sizeHint()
        self.button.setMaximumWidth(size.height())
        self._layout.insertWidget(self._layout.count() - 1, self.button)

    def color_min(self) -> float:
        return self._color_min

    def set_color_min(self, color_min: float) -> None:
        self._color_min = color_min
        for line in self.lines:
            line.set_minimum(self._color_min)  # noqa

    def color_max(self) -> float:
        return self._color_max

    def set_color_max(self, color_max: float) -> None:
        self._color_max = color_max
        for line in self.lines:
            line.set_maximum(self._color_max)  # noqa

    def select_color(self) -> None:
        options = QtWidgets.QColorDialog.ColorDialogOption.DontUseNativeDialog
        color = QtWidgets.QColorDialog.getColor(initial=self._value, options=options)
        if color.isValid():
            super().set_value(color)  # noqa
            values = self._cast_to_tuple(color)
            self._set_line_values(values)  # noqa
            self._set_button_value(color)

    def value(self) -> QtGui.QColor:
        return super().value()  # noqa

    def set_value(self, value: QtGui.QColor | Sequence) -> None:
        super().set_value(value)  # noqa
        self._set_button_value(self._value)

    def _cast_to_type(self, values: tuple[float, ...]) -> QtGui.QColor:
        return QtGui.QColor.fromRgbF(*values[:3])

    def _cast_to_tuple(self, value: QtGui.QColor) -> tuple[float, ...]:
        return value.getRgbF()[:3]

    def _line_value_changed(self, value: float) -> None:
        super()._line_value_changed(value)  # noqa
        self._set_button_value(self._value)

    def _set_button_value(self, value: QtGui.QColor) -> None:
        self.button.setPalette(QtGui.QPalette(value))

    def _slider_value_changed(self, value: int) -> None:
        super()._slider_value_changed(value)
        self._set_button_value(self._value)



from .box import CollapsibleBox

from .tabdata import TabDataParameter

from .widgets import (
    ParameterWidget,
    IntParameter,
    FloatParameter,
    StringParameter,
    PathParameter,
    ComboParameter,
    EnumParameter,
    BoolParameter,
    MultiIntParameter,
    MultiFloatParameter,
    PointParameter,
    PointFParameter,
    SizeParameter,
    SizeFParameter,
    ColorParameter,
)

from .inputs import Label


from .list import (
    ListParameter,
    StringListParameter,
)

from .editor import (
    ParameterToolTip,
    ParameterLabel,
    ParameterForm,
    ParameterEditor,
)


__version__ = '1.3.1'



from __future__ import annotations

import logging
import math
from numbers import Number
from typing import Generic, TypeVar

from qt_material_icons import MaterialIcon
from qtpy import QtCore, QtGui, QtWidgets

SUCCESS = 25


N = TypeVar('N', bound=Number)


class NumberLineEdit(QtWidgets.QLineEdit, Generic[N]):
    value_changed = QtCore.Signal(int)

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._init_validator()

        self._abs_minimum = self._validator.bottom()
        self._abs_maximum = self._validator.top()
        self._minimum = self._abs_minimum
        self._maximum = self._abs_maximum
        self._value = 0

        self.commit_on_edit = False

        self.editingFinished.connect(self.commit)
        self.textEdited.connect(self._text_edit)

    def _init_validator(self) -> None:
        self._validator = IntValidator()
        self.setValidator(self._validator)

    def value(self) -> N:
        return self._value

    def set_value(self, value: N) -> None:
        text = self._validator.fixup(str(value))
        state, text_, pos_ = self._validator.validate(text, 0)
        if state == QtGui.QValidator.State.Acceptable:
            self.setText(text)
            self.commit()

    def minimum(self) -> N:
        return self._minimum

    def set_minimum(self, minimum: N | None) -> None:
        if minimum is None:
            minimum = self._abs_minimum
        self._minimum = minimum
        self._validator.setBottom(minimum)

    def maximum(self) -> N:
        return self._maximum

    def set_maximum(self, maximum: N | None) -> None:
        if maximum is None:
            maximum = self._abs_maximum
        self._maximum = maximum
        self._validator.setTop(maximum)

    def commit(self, update_text: bool = True) -> None:
        """Commit the current text."""

        try:
            value = int(self.text())
        except ValueError:
            value = 0

        # Strip padding
        if int(value) == value:
            value = int(value)
        if value != self._value:
            self._value = value
            self.value_changed.emit(value)
        if update_text:
            self.setText(str(value))

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        if event.key() == QtCore.Qt.Key.Key_Up:
            self._step(add=True)
            event.accept()
        elif event.key() == QtCore.Qt.Key.Key_Down:
            self._step(add=False)
            event.accept()
        else:
            super().keyPressEvent(event)

    def minimumSizeHint(self) -> QtCore.QSize:
        size = super().minimumSizeHint()
        size.setWidth(24)
        return size

    def sizeHint(self) -> QtCore.QSize:
        size = super().sizeHint()
        size.setWidth(60)
        return size

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        delta = event.angleDelta()
        if delta.y() > 0:
            self._step(add=True)
        elif delta.y() < 0:
            self._step(add=False)
        event.accept()

    def _step(self, add: bool) -> bool:
        """Step up or down the text based on the current cursor position."""

        self.setFocus()
        text = self.text() or '0'
        position = self.cursorPosition()
        if self.hasSelectedText():
            position = self.selectionStart()

        # Check if the cursor is on a special character
        if position < len(text) and not text[position].isdigit():
            return False

        step_index = self._step_index(text, position)
        exponent = self._step_exponent(step_index)

        # Perform a step up or down
        amount = 1 if add else -1
        step = amount * pow(10, exponent)
        value = self._value + step

        text = self._format_value(value, text, exponent)

        # Validate before setting new text
        state, text_, pos_ = self.validator().validate(text, 0)
        if state != QtGui.QValidator.State.Acceptable:
            return False
        self.setText(text)

        # Don't commit change to preserve padding
        self._value = value
        self.value_changed.emit(value)

        # Get the new position and set selection
        position = self._relative_position(step_index, text)
        self.setSelection(position, 1)
        return True

    def _text_edit(self) -> None:
        """Handle when the text has been edited."""

        if self.commit_on_edit:
            self.commit(update_text=False)

    def _format_value(self, value: int, current_text: str, exponent: int) -> str:
        """Return the text for a value, preserving the format of the current text."""

        padding = len([t for t in current_text if t.isdigit()])
        if value < 0:
            padding += 1
        text = f'{value:0{padding}}'
        return text

    def _step_exponent(self, step_index: int) -> int:
        """Return the exponent for the cursor position."""

        exponent = step_index - 1
        return exponent

    def _step_index(self, text: str, position: int) -> int:
        """
        Return the step index relative to the decimal point.
        This preserves position when the number gets larger or changes +/- sign.
        """

        step_index = len(text) - position
        # If cursor is at end, edit the first digit
        step_index = max(1, step_index)

        return step_index

    def _relative_position(self, step_index: int, text: str) -> int:
        """Return the relative position for the text at step index."""

        position = len(text) - step_index
        return position


class IntLineEdit(NumberLineEdit[int]):
    value_changed = QtCore.Signal(int)


class FloatLineEdit(NumberLineEdit[float]):
    value_changed = QtCore.Signal(float)

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)
        self._decimals = self._validator.decimals()

    def _init_validator(self) -> None:
        self._validator = DoubleValidator()
        # NOTE: Using QLocale.c() fixes validation issues in non-English locales that
        # use comma decimal separators.
        self._validator.setLocale(QtCore.QLocale.c())
        self._validator.setNotation(QtGui.QDoubleValidator.Notation.StandardNotation)
        self.setValidator(self._validator)

    def commit(self, update_text: bool = True) -> None:
        """Commit the current text."""

        try:
            value = float(self.text())
        except ValueError:
            value = float(0)

        # Strip padding
        if value != self._value:
            self._value = value
            self.value_changed.emit(value)
        if update_text:
            if int(value) == value:
                value = int(value)
            self.setText(str(value))

    def decimals(self) -> int:
        return self._decimals

    def set_decimals(self, value: int) -> None:
        self._decimals = value
        self._validator.setDecimals(value)

    def _format_value(self, value: int, current_text: str, exponent: int) -> str:
        decimal_index = current_text.find('.')

        # Preserve padding
        if decimal_index == -1:
            padding_decimal = 0
        else:
            padding_decimal = len(current_text) - 1 - decimal_index
            current_text = current_text[:decimal_index]

        # Preserve padding when switching to something like 1.001 > 1.000
        padding_decimal = max(padding_decimal, -exponent)
        padding_int = len([t for t in current_text if t.isdigit()])
        # Account for minus sign
        if value < 0:
            padding_int += 1

        # padding_int needs to contain both padding for int and decimals
        padding_int += padding_decimal + 1 * bool(padding_decimal)

        value = round(value, padding_decimal)
        text = f'{value:0{padding_int}.{padding_decimal}f}'

        return text

    def _step_exponent(self, step_index: int) -> int:
        exponent = step_index
        # If the cursor is on a decimal edit the first decimal.
        if step_index >= 0:
            exponent = step_index - 1

        return exponent

    def _step_index(self, text: str, position: int) -> int:
        decimal_index = text.find('.')
        if decimal_index == -1:
            step_index = len(text) - position
        else:
            step_index = decimal_index - position
        return step_index

    def _relative_position(self, step_index: int, text: str) -> int:
        decimal_index = text.find('.')
        position = len(text) - step_index
        if decimal_index > -1:
            # If the position is on a decimal point, move to the first decimal.
            if step_index == 0:
                step_index = -1
            position = decimal_index - step_index
        return position


class IntValidator(QtGui.QIntValidator):
    def fixup(self, text: str) -> str:
        text = str(super().fixup(text))
        text = text.replace(',', '')
        try:
            text = str(max(min(int(text), self.top()), self.bottom()))
        except ValueError:
            pass
        return text


class DoubleValidator(QtGui.QDoubleValidator):
    def fixup(self, text: str) -> str:
        try:
            float(text)
        except ValueError:
            characters = '+-01234567890.'
            text = ''.join(t for t in text if t in characters)

        try:
            value = float(text)
            value = min(max(value, self.bottom()), self.top())
            value = round(value, self.decimals())
            text = '{value:.{decimals}f}'.format(value=value, decimals=self.decimals())
            return text
        except (ValueError, TypeError):
            return text


class NumberSlider(QtWidgets.QSlider, Generic[N]):
    value_changed = QtCore.Signal(Number)

    def __init__(
        self,
        orientation: QtCore.Qt.Orientation = QtCore.Qt.Orientation.Horizontal,
        parent: QtWidgets.QWidget | None = None,
    ) -> None:
        super().__init__(orientation, parent)

        self._minimum = super().minimum()
        self._maximum = super().maximum()
        self._step_factor = 2

        self.setTickPosition(QtWidgets.QSlider.TickPosition.TicksBothSides)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

        self.valueChanged.connect(self._value_changed)

    def value(self) -> N:
        value = super().value()
        real_value = self._real_value(value)
        return real_value

    def set_value(self, value: N) -> None:
        if math.isnan(value):
            return
        slider_value = self._slider_value(value)
        self.setSliderPosition(slider_value)


    def minimum(self) -> float:
        return self._minimum

    def set_minimum(self, minimum: float) -> None:
        value = self.value()
        self._minimum = minimum
        self._refresh_steps()
        self.set_value(value)

    def maximum(self) -> float:
        return self._maximum

    def set_maximum(self, maximum: float) -> None:
        value = self.value()
        self._maximum = maximum
        self._refresh_steps()
        self.set_value(value)

    def step_factor(self) -> int:
        return self._step_factor

    def set_step_factor(self, factor: int) -> None:
        """
        Set the step factor for the slider. The higher the factor, the smaller the step
        size and page size.
        """

        self._step_factor = factor
        self._refresh_steps()

    def _slider_value(self, value: N) -> int:
        """Return the value for the slider from the 'real' value."""

        try:
            percentage = (value - self._minimum) / (self._maximum - self._minimum)
        except ZeroDivisionError:
            return 0
        slider_range = super().maximum() - super().minimum()
        clamped_value = min(max(percentage, 0), 1) * slider_range + super().minimum()
        return int(clamped_value)

    def _real_value(self, value: int) -> N:
        """Return the 'real' value from the slider value."""

        slider_range = super().maximum() - super().minimum()
        try:
            percentage = (value - super().minimum()) / slider_range
        except ZeroDivisionError:
            return float('nan')
        real_value = self._minimum + (self._maximum - self._minimum) * percentage
        return real_value


    def _exponent(self) -> int:
        """Return the exponent based on the minimum and maximum."""

        num_range = abs(self.maximum() - self.minimum())
        if num_range == 0:
            num_range = 1
        exponent = math.log10(num_range)

        # Round exponent up or down with weighting towards down
        if exponent % 1 > 0.8:
            exponent = math.ceil(exponent)
        else:
            exponent = math.floor(exponent)
        return exponent


    def _refresh_steps(self) -> None:
        """Refresh the slider ticks and steps based on the minimum and maximum."""

        value = self.value()

        page_step = max(10, pow(10, self._step_factor - 1))
        self.setSingleStep(1)
        self.setPageStep(page_step)
        self.setTickInterval(page_step)

        factor = self._slider_factor()

        self.blockSignals(True)
        self.setMinimum(int(self._minimum * factor))
        self.setMaximum(int(self._maximum * factor))
        self.set_value(value)
        self.blockSignals(False)

    def _slider_factor(self) -> float:
        """
        Return the scale factor for the slider to convert the real range into the
        slider range.
        """

        factor = pow(10, -(self._exponent() - self._step_factor))
        return factor

    def _value_changed(self, value: int) -> None:
        """Emit a signal on value change with the real value."""

        value = self._real_value(value)
        if not math.isnan(value):
            self.value_changed.emit(value)

class IntSlider(NumberSlider[int]):
    value_changed = QtCore.Signal(int)

    def _slider_factor(self) -> float:
        factor = pow(10, -(self._exponent() - self._step_factor))
        factor = min(1, factor)
        return factor

class FloatSlider(NumberSlider[float]):
    value_changed = QtCore.Signal(float)


class RatioButton(QtWidgets.QPushButton):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent=parent)

        self._icon_off = MaterialIcon('link_off')
        self._icon_on = MaterialIcon('link')

        self.toggled.connect(self._checked_change)
        size = self.iconSize().width()
        self.setMaximumSize(QtCore.QSize(size, size))
        self.setCheckable(True)
        self._checked_change(False)

    def _checked_change(self, checked: bool) -> None:
        # BUG: fusion style does not recognize On/Off for QIcons
        # https://bugreports.qt.io/browse/QTBUG-82110
        icon = self._icon_on if checked else self._icon_off
        super().setIcon(icon)


class TextEdit(QtWidgets.QPlainTextEdit):
    editing_finished = QtCore.Signal()

    def focusOutEvent(self, event: QtGui.QFocusEvent) -> None:
        self.editing_finished.emit()
        return super().focusOutEvent(event)

    def sizeHint(self) -> QtCore.QSize:
        size_hint = super().sizeHint()
        size_hint.setHeight(self.minimumSizeHint().height())
        return size_hint


class Label(QtWidgets.QWidget):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._icon = None
        style = self.style()
        icon_size = style.pixelMetric(QtWidgets.QStyle.PixelMetric.PM_ButtonIconSize)
        self._icon_size = QtCore.QSize(icon_size, icon_size)

        self._init_ui()

    def _init_ui(self) -> None:
        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(QtCore.QMargins())
        self.setLayout(layout)

        self._icon_label = QtWidgets.QLabel()
        layout.addWidget(self._icon_label)
        self._text_label = QtWidgets.QLabel()
        self._text_label.setWordWrap(True)
        layout.addWidget(self._text_label)
        layout.setStretch(1, 1)

    def icon(self) -> QtGui.QIcon | None:
        return self._icon

    def set_icon(self, icon: QtGui.QIcon | None) -> None:
        self._icon = icon
        self._refresh_icon()

    def icon_size(self) -> QtCore.QSize:
        return self._icon_size

    def set_icon_size(self, icon_size: QtCore.QSize) -> None:
        self._icon_size = icon_size
        self._refresh_icon()

    def text(self) -> str:
        return self._text_label.text()

    def set_text(self, text: str) -> None:
        self._text_label.setText(text)

    def set_level(self, level: int) -> None:
        icon = None
        color = None
        if level >= logging.CRITICAL:
            icon = MaterialIcon('report')
            color = self._color('magenta')
        elif level >= logging.ERROR:
            icon = MaterialIcon('error')
            color = self._color('red')
        elif level >= logging.WARNING:
            icon = MaterialIcon('warning')
            color = self._color('orange')
        elif level >= SUCCESS:
            icon = MaterialIcon('check_circle')
            color = self._color('green')
        elif level >= logging.INFO:
            icon = MaterialIcon('info')

        self.set_icon(icon)
        if icon:
            # Create a custom pixmap with color.
            self._icon_label.setPixmap(icon.pixmap(size=self._icon_size, color=color))

    def _refresh_icon(self) -> None:
        if self._icon:
            self._icon_label.setPixmap(self._icon.pixmap(self._icon_size))
        else:
            self._icon_label.clear()

    @staticmethod
    def _color(name: str) -> QtGui.QColor | None:
        try:
            import qt_themes
        except ImportError:
            return None

        if theme := qt_themes.get_theme():
            return getattr(theme, name, None)
        return None
